#include <windows.h>
#include <stdio.h>
#include "stdserv.h"

#ifdef __GNUC__
#include <string.h>
#define _stricmp strcasecmp
#endif

#define PATHLEN 1024

DWORD                    serviceCurrentStatus;
SERVICE_STATUS_HANDLE    serviceStatusHandle;
HANDLE                   killServiceEvent = NULL;
HANDLE                   threadHandle = NULL;
BOOL                     serviceRunning = FALSE;
BOOL                     servicePaused = FALSE;
BOOL                     DebugMode=FALSE;

VOID ServiceMain(DWORD argc, LPTSTR *argv);
void terminateService(DWORD ErrorNumber);
BOOL StartServiceThread(void);
VOID TerminateServiceThread(void);
BOOL UpdateSCMStatus(DWORD dwCurrentState,
                     DWORD dwWin32ExitCode,
                     DWORD dwServiceSpecificExitCode,
                     DWORD dwCheckPoint,
                     DWORD dwWaitHint);
VOID ServiceCtrlHandler(DWORD controlCode);
void KillService(void);
void HandleError(char *ErrorText, DWORD ErrorNumber);
void InstallService(void);
void RemoveService(void);
void Usage(char *a0);


// ServiceMain -
// ServiceMain is called when the Service Control Manager wants to
// launch the service.  It should not return until the service has
// stopped. To accomplish this, for this example, it waits blocking
// on an event just before the end of the function.  That event gets
// set by the function which terminates the service above.  Also, if
// there is an error starting the service, ServiceMain returns immediately
// without launching the main service thread, terminating the service.
VOID ServiceMain(DWORD argc, LPTSTR *argv)
{
    BOOL success;
    // First we must call the Registration function
    serviceStatusHandle = RegisterServiceCtrlHandler(SERVICE_NAME,
                                                     (LPHANDLER_FUNCTION) ServiceCtrlHandler);
    if (!serviceStatusHandle) {
        terminateService(GetLastError());
        return;
    }
    // Next Notify the Service Control Manager of progress
    success = UpdateSCMStatus(SERVICE_START_PENDING, NO_ERROR, 0, 1, 5000);
    if (!success) {
        terminateService(GetLastError());
        return;
    }
    // Now create the our service termination event to block on
    killServiceEvent = CreateEvent (0, TRUE, FALSE, 0);
    if (!killServiceEvent) {
        terminateService(GetLastError());
        return;
    }
    // Notify the SCM of progress again
    success = UpdateSCMStatus(SERVICE_START_PENDING, NO_ERROR, 0, 2, 3000);
    if (!success) {
        terminateService(GetLastError());
        return;
    }
    // Do some thread initialization (e.g. set up a stop event)
    InitServiceThread(); // SERVICE SPECIFIC FUNCTION //
    // Notify the SCM of progress again...
    success = UpdateSCMStatus(SERVICE_START_PENDING, NO_ERROR, 0, 3, 5000);
    if (!success) {
        terminateService(GetLastError());
        return;
    }
    // Start the service execution thread by calling our StartServiceThread function...
    success = StartServiceThread();
    if (!success) {
        TerminateServiceThread();
        terminateService(GetLastError());
        return;
    }
    // The service is now running.  Notify the SCM of this fact.
    serviceCurrentStatus = SERVICE_RUNNING;
    success = UpdateSCMStatus(SERVICE_RUNNING, NO_ERROR, 0, 0, 0);
    if (!success) {
        terminateService(GetLastError());
        return;
    }
    // Now just wait for our killed service signal, and then exit, which
    // terminates the service!
    WaitForSingleObject (killServiceEvent, INFINITE);
    Sleep(500);
    TerminateServiceThread();
    terminateService(0);
}

void terminateService(DWORD ErrorNumber)
{
    // optional cleanup
    if (ErrorNumber!=0) {
        printf("Fatal error #%i, terminating.\n",ErrorNumber);
    }
    UpdateSCMStatus(SERVICE_STOPPED, ErrorNumber, 0, 0, 0);
}

// StartServiceThread -
// This starts the service by creating its execution thread.
BOOL StartServiceThread(void)
{
    DWORD id;
    // Start the service's thread
    threadHandle = CreateThread(0,
                                0,
                                (LPTHREAD_START_ROUTINE) ServiceExecutionThread, // SERVICE SPECIFIC FUNCTION //
                                0,
                                0,
                                &id);
    if (threadHandle == 0) {
        return FALSE;
    } else {
        serviceRunning = TRUE;
        return TRUE;
    }
}

// TerminateServiceThread -
// The service thread should be ended before the service returns to the SCM.
// Either it should honor the serviceRunning flag, or it should be stopped
// with the user-supplied function StopServiceThread. At a last resort,
// after 500ms grace, the thread is terminated.
VOID TerminateServiceThread(void)
{
    DWORD dwExitCode=0;

    StopServiceThread(); // SERVICE SPECIFIC FUNCTION //
    Sleep(500);
    GetExitCodeThread(threadHandle,&dwExitCode);
    if (dwExitCode==STILL_ACTIVE) {
        TerminateThread(threadHandle,0);
    }
}

// This function updates the service status for the SCM
BOOL UpdateSCMStatus (DWORD dwCurrentState,
                      DWORD dwWin32ExitCode,
                      DWORD dwServiceSpecificExitCode,
                      DWORD dwCheckPoint,
                      DWORD dwWaitHint)
{
    BOOL success;
    SERVICE_STATUS serviceStatus;
    // Fill in all of the SERVICE_STATUS fields
    serviceStatus.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
    serviceStatus.dwCurrentState = dwCurrentState;
    // If in the process of something, then accept
    // no control events, else accept anything
    if (dwCurrentState == SERVICE_START_PENDING) {
        serviceStatus.dwControlsAccepted = 0;
    } else {
        serviceStatus.dwControlsAccepted = SERVICE_ACCEPT_STOP |
                                           SERVICE_ACCEPT_PAUSE_CONTINUE |
                                           SERVICE_ACCEPT_SHUTDOWN;
    }
    // if a specific exit code is defines, set up
    // the Win32 exit code properly
    if (dwServiceSpecificExitCode == 0) {
        serviceStatus.dwWin32ExitCode = dwWin32ExitCode;
    } else {
        serviceStatus.dwWin32ExitCode = ERROR_SERVICE_SPECIFIC_ERROR;
    }
    serviceStatus.dwServiceSpecificExitCode = dwServiceSpecificExitCode;
    serviceStatus.dwCheckPoint = dwCheckPoint;
    serviceStatus.dwWaitHint = dwWaitHint;
    // Pass the status record to the SCM
    success = SetServiceStatus (serviceStatusHandle, &serviceStatus);
    if (!success) {
        KillService();
    }
    return success;
}

// Handles the events dispatched by the Service Control Manager.
VOID ServiceCtrlHandler (DWORD controlCode)
{
    BOOL success;
    switch (controlCode) {
    // There is no START option because
    // ServiceMain gets called on a start
    // Pause the service
    case SERVICE_CONTROL_PAUSE:
        if (serviceRunning && !servicePaused) {
            // Tell the SCM we're about to Pause.
            success = UpdateSCMStatus(SERVICE_PAUSE_PENDING, NO_ERROR,
                                      0, 1, 1000);
            servicePaused = TRUE;
            SuspendThread(threadHandle);
            serviceCurrentStatus = SERVICE_PAUSED;
        }
        break;
    // Resume from a pause
    case SERVICE_CONTROL_CONTINUE:
        if (serviceRunning && servicePaused) {
            // Tell the SCM we're about to Resume.
            success = UpdateSCMStatus(SERVICE_CONTINUE_PENDING, NO_ERROR, 0,
                                      1, 1000);
            servicePaused=FALSE;
            ResumeThread(threadHandle);
            serviceCurrentStatus = SERVICE_RUNNING;
        }
        break;
    // Update the current status for the SCM.
    case SERVICE_CONTROL_INTERROGATE:
        // This does nothing, here we will just fall through to the end
        // and send our current status.
        break;
    // For a shutdown, we can do cleanup but it must take place quickly
    // because the system will go down out from under us.
    // For this app we have time to stop here, which I do by just falling
    // through to the stop message.
    case SERVICE_CONTROL_SHUTDOWN:
    // Stop the service
    case SERVICE_CONTROL_STOP:
        // Tell the SCM we're about to Stop.
        serviceCurrentStatus = SERVICE_STOP_PENDING;
        success = UpdateSCMStatus(SERVICE_STOP_PENDING, NO_ERROR, 0, 1, 5000);
        KillService();
        return;
    default:
        break;
    }
    UpdateSCMStatus(serviceCurrentStatus, NO_ERROR, 0, 0, 0);
}

// the killServiceEvent to allow ServiceMain to exit.
void KillService(void)
{
    serviceRunning=FALSE;
    // Set the event that is blocking ServiceMain
    // so that ServiceMain can return
    SetEvent(killServiceEvent);
}

void HandleError(char *ErrorText, DWORD ErrorNumber)
{
    printf("Fatal error #%i: %s\n",ErrorNumber,ErrorText);
    exit(1);
}

void InstallService(void)
{
    SC_HANDLE myService, scm;
    char szFullPath[PATHLEN];

    if (GetModuleFileName(NULL,szFullPath,PATHLEN) == 0) {
        HandleError("Installation failed, cannot get path to module.",GetLastError());
    }

    printf("Installing Service...\n");
    // open a connection to the SCM
    scm = OpenSCManager(0, 0, SC_MANAGER_CREATE_SERVICE);
    if (!scm) {
        HandleError("Installation failed, cannot open SCManager.", GetLastError());
    }
    printf("Opened Service Control Manager...\n");
    // Install the new service
    myService = CreateService(scm,
                              SERVICE_NAME,                  // the internal service name used by the SCM
                              CP_SERVICE_NAME,              // the external label seen in the Service Control applet
                              SERVICE_ALL_ACCESS,          // We want full access to control the service
                              SERVICE_WIN32_OWN_PROCESS, // The service is a single app and not a driver
                              SERVICE_DEMAND_START,        // The service will be started by us manually
                              SERVICE_ERROR_NORMAL,        // If error during service start, don't misbehave.
                              szFullPath,
                              0, 0, 0, 0, 0);
    if (!myService) {
        HandleError("Installation failed, cannot create service.", GetLastError());
    } else {
        printf("Service successfully installed.\n");
    }
    // clean up
    CloseServiceHandle(myService);
    CloseServiceHandle(scm);
}

void RemoveService(void)
{
    SC_HANDLE myService, scm;
    BOOL success;
    SERVICE_STATUS status;
    printf("Removing Service...\n");
    // Open a Service Control Manager connection
    scm = OpenSCManager(0, 0, SC_MANAGER_CREATE_SERVICE);
    if (!scm) {
        HandleError("Remove failed, cannot open SCManager.", GetLastError());
    }
    printf("Opened Service Control Manager...\n");
    // Get the service's handle
    myService = OpenService(scm, SERVICE_NAME, SERVICE_ALL_ACCESS | DELETE);
    if (!myService) {
        HandleError("Remove failed, cannot open service.", GetLastError());
    }
    // Stop the service if necessary
    success = QueryServiceStatus(myService, &status);
    if (!success) {
        HandleError("Remove failed, cannot query service status.", GetLastError());
    }
    if (status.dwCurrentState != SERVICE_STOPPED) {
        printf("Service currently active.  Stopping service...\n");
        success = ControlService(myService, SERVICE_CONTROL_STOP, &status);
        if (!success) {
            HandleError("ControlService fails to stop service!", GetLastError());
        }
        Sleep(500);
    }
    // Remove the service
    success = DeleteService(myService);
    if (success) {
        printf("Service successfully removed.\n");
    } else {
        HandleError("Remove failed, cannot delete service.", GetLastError());
    }
    CloseServiceHandle(myService);
    CloseServiceHandle(scm);
}

void Usage(char *a0)
{
    printf("Usage: %s [-install | -remove"
#ifdef DEBUGCMDLINEOPT
           " | -now"
#endif
           "]\n",a0);
    printf("         -install         install the service\n");
    printf("         -remove          remove the service\n");
#ifdef DEBUGCMDLINEOPT
    printf("         -%-15s %s\n",DEBUGCMDLINEOPT,DEBUGUSAGETEXT);
#endif
    printf("\nDo not call without parameters, this is reserved for the SCM!\n");
}

int main(int argc, char *argv[])
{
    SERVICE_TABLE_ENTRY serviceTable[] = {
        {SERVICE_NAME, (LPSERVICE_MAIN_FUNCTION) ServiceMain},
        {NULL, NULL} };
    BOOL success;
    
    if ((argc > 1) && ((*argv[1] == '-')||(*argv[1] == '/'))) {
        if (_stricmp("install",argv[1]+1) == 0) {
            InstallService();
            exit(0);
        } else if (_stricmp("remove",argv[1]+1) == 0) {
            RemoveService();
            exit(0);
#ifdef DEBUGCMDLINEOPT
        } else if (_stricmp(DEBUGCMDLINEOPT,argv[1]+1) == 0) {
            DebugMode=TRUE;
            RunServiceInDebugMode(--argc, ++argv); // SERVICE SPECIFIC FUNCTION //
            exit(0);
#endif
        } else {
            Usage(argv[0]);
            exit(1);
        }
    }
    Usage(argv[0]); // print usage message: if started from SCM, this has no effect
    printf("If have done it nevertheless, you will see an error message #1063 soon...\n");
    // Register the service with the Service Control Manager
    success = StartServiceCtrlDispatcher(serviceTable);
    if (!success) {
        // This error message would not show up unless we had the "allow service to
        // interact with desktop" option checked, which is unlikely for a console mode
        // app, but we'll print an error message anyway just in case there's someone
        // around to see it.
        HandleError("StartServiceCtrlDispatcher fails!", GetLastError());
    }
    return 0;
}

